from policyengine_uk.model_api import *


class non_primary_residence_wealth_tax(Variable):
    label = "Wealth tax (non-primary residence)"
    documentation = (
        "Annual tax on household net wealth excluding primary residences"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        wealth = household("total_wealth", period)
        primary_residence = household("main_residence_value", period)
        tax = parameters(period).gov.contrib.ubi_center.wealth_tax
        return tax.calc(max_(0, wealth - primary_residence))



from policyengine_uk.model_api import *


class pension_credit_entitlement(Variable):
    label = "PC entitlement"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    defined_for = "is_pension_credit_eligible"
    adds = ["guarantee_credit", "savings_credit"]



from policyengine_uk.model_api import *


class child_benefit_reported(Variable):
    label = "Child Benefit (reported amount)"
    documentation = "Reported amount received for Child Benefit"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP


class would_claim_child_benefit(Variable):
    label = "Would claim Child Benefit"
    documentation = (
        "Whether this benefit unit would claim Child Benefit if eligible"
    )
    entity = BenUnit
    definition_period = YEAR
    value_type = bool

    def formula(benunit, period, parameters):
        claims_benefits = benunit("claims_all_entitled_benefits", period)
        already_claiming = add(benunit, period, ["child_benefit_reported"]) > 0
        takeup_rate = parameters(period).gov.hmrc.child_benefit.takeup
        baseline_cb = benunit("baseline_child_benefit_entitlement", period) > 0
        eligible = benunit("child_benefit_entitlement", period) > 0
        return select(
            [already_claiming | claims_benefits, ~baseline_cb & eligible],
            [
                True,  # Claims Child Benefit in the baseline
                random(benunit)
                < takeup_rate,  # New CB eligibility from a reform
            ],
            default=False,  # Always non-claimant
        )


class child_benefit_respective_amount(Variable):
    label = "Child Benefit (respective amount)"
    documentation = "The amount of this benefit unit's Child Benefit which is in respect of this person"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = (
        "https://www.legislation.gov.uk/ukpga/1992/4/part/IX",
        "https://www.legislation.gov.uk/uksi/2006/965/regulation/2",
    )
    defined_for = "is_child_or_QYP"

    def formula(person, period, parameters):
        eligible = True
        if parameters(
            period
        ).gov.contrib.ubi_center.basic_income.interactions.withdraw_cb:
            eligible &= person("basic_income", period) == 0
        is_eldest = person("is_eldest_child", period)
        child_benefit = parameters(period).gov.hmrc.child_benefit.amount
        amount = where(
            is_eldest, child_benefit.eldest, child_benefit.additional
        )
        return eligible * amount * WEEKS_IN_YEAR


class child_benefit_entitlement(Variable):
    label = "CB entitlement"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP

    adds = ["child_benefit_respective_amount"]


class child_benefit(Variable):
    label = "Child Benefit"
    documentation = "Total Child Benefit for the benefit unit"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    category = BENEFIT
    defined_for = "would_claim_child_benefit"
    adds = ["child_benefit_entitlement"]


class child_benefit_less_tax_charge(Variable):
    label = "Child Benefit (less tax charge)"
    documentation = (
        "Child Benefit, minus the Child Benefit High-Income Tax Charge"
    )
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP

    adds = ["child_benefit"]
    subtracts = ["CB_HITC"]


class baseline_child_benefit_entitlement(Variable):
    label = "Child Benefit (baseline)"
    entity = BenUnit
    definition_period = YEAR
    value_type = float



from policyengine_uk.model_api import *

MICRODATA_VAT_COVERAGE = 0.38
"""
LCFS (from which ETB data is derived) microdata is known to under-report household consumption. We scale up the VAT liability to hit HMRC-reported VAT receipts (following the approach of IFS' TAXBEN, though they might have better coverage anyway from access to the non-EUL dataset).

For HMRC statistics see: https://www.gov.uk/government/statistics/value-added-tax-vat-annual-statistics
"""


class vat(Variable):
    label = "VAT"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        full_rate_consumption = household("full_rate_vat_consumption", period)
        reduced_rate_consumption = household(
            "reduced_rate_vat_consumption", period
        )
        vat = parameters(period).gov.hmrc.vat
        return (
            full_rate_consumption * vat.standard_rate
            + reduced_rate_consumption * vat.reduced_rate
        ) / MICRODATA_VAT_COVERAGE


class baseline_vat(Variable):
    label = "baseline VAT"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        full_rate_consumption = household("full_rate_vat_consumption", period)
        reduced_rate_consumption = household(
            "reduced_rate_vat_consumption", period
        )
        vat = parameters(period).baseline.gov.hmrc.vat
        return (
            full_rate_consumption * vat.standard_rate
            + reduced_rate_consumption * vat.reduced_rate
        ) / MICRODATA_VAT_COVERAGE


class vat_change(Variable):
    label = "change in VAT liability"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"
    adds = ["vat"]
    subtracts = ["baseline_vat"]



# Run this script to add VAT imputation to the enhanced FRS.

import pandas as pd
from microdf import MicroDataFrame
from synthimpute import rf_impute
from policyengine_uk import Microsimulation

CONSUMPTION_PCT_REDUCED_RATE = 0.03  # From OBR's VAT page
CURRENT_VAT_RATE = 0.2

etb = pd.read_csv(
    "~/Downloads/UKDA-8856-tab/tab/householdv2_1977-2021.tab", delimiter="\t"
)
etb_2020 = etb[etb.year == 2020].dropna()
for col in etb_2020:
    etb_2020[col] = pd.to_numeric(etb_2020[col], errors="coerce")

sim = Microsimulation()


INPUT_VARIABLES = ["is_adult", "is_child", "is_SP_age", "household_net_income"]

frs_df = sim.calculate_dataframe(
    INPUT_VARIABLES[::-1],
    map_to="household",
)

imputed_rates = rf_impute(
    x_train=etb_2020_df.drop("full_rate_vat_expenditure_rate", axis=1),
    y_train=etb_2020_df.full_rate_vat_expenditure_rate,
    x_new=frs_df[INPUT_VARIABLES],
)

etb_2020_df = pd.DataFrame()
etb_2020_df["is_adult"] = etb_2020.adults
etb_2020_df["is_child"] = etb_2020.childs
etb_2020_df["is_SP_age"] = etb_2020.noretd
etb_2020_df["household_net_income"] = etb_2020.disinc * 52
etb_2020_df["full_rate_vat_expenditure_rate"] = (
    etb_2020.totvat * (1 - CONSUMPTION_PCT_REDUCED_RATE) / CURRENT_VAT_RATE
) / (etb_2020.expdis - etb_2020.totvat)
etb_2020_df = etb_2020_df[~etb_2020_df.full_rate_vat_expenditure_rate.isna()]
etb_2020_df = MicroDataFrame(
    etb_2020_df, weights=etb_2020.HHold_unadj_weight * 1_000
)
non_reduced_rate_consumption = sim.calculate(
    "consumption", 2022
) - sim.calculate("reduced_rate_vat_consumption", 2022)
sim.set_input(
    "full_rate_vat_consumption",
    2022,
    imputed_rates * non_reduced_rate_consumption,
)

EnhancedFRS.save(
    2022,
    "full_rate_vat_consumption/2022",
    sim.calculate("full_rate_vat_consumption", 2022).values,
)



from policyengine_uk.model_api import *


class monthly_domestic_energy_consumption(Variable):
    label = "Monthly domestic energy consumption"
    entity = Household
    definition_period = MONTH
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        # From RF "A chilling crisis", p11. Available at https://www.resolutionfoundation.org/app/uploads/2022/08/A-chilling-crisis.pdf
        # This is very approximate but gives us some distribution of seasonal energy consumption at least.
        ENERGY_CONSUMPTION_RATIOS = [
            350,
            350,
            350,
            180,
            180,
            180,
            180,
            180,
            180,
            300,
            300,
            300,
        ]
        consumption_distribution = np.array(ENERGY_CONSUMPTION_RATIOS) / sum(
            ENERGY_CONSUMPTION_RATIOS
        )
        month = period.start.month
        return (
            household("domestic_energy_consumption", period.this_year)
            * consumption_distribution[month - 1]
        )


class monthly_epg_consumption_level(Variable):
    label = "Monthly EPG subsidy level"
    entity = Household
    definition_period = MONTH
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        energy_consumption = household(
            "monthly_domestic_energy_consumption", period
        )
        ofgem = parameters.gov.ofgem
        price_cap = ofgem.energy_price_cap(period)
        price_guarantee = ofgem.energy_price_guarantee(period)
        return energy_consumption * price_guarantee / price_cap


class monthly_epg_subsidy(Variable):
    label = "Monthly EPG subsidy"
    entity = Household
    definition_period = MONTH
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        energy_consumption = household(
            "monthly_domestic_energy_consumption", period
        )
        epg_consumption_level = household(
            "monthly_epg_consumption_level", period
        )
        return max_(0, energy_consumption - epg_consumption_level)


class epg_subsidy(Variable):
    label = "Energy price guarantee subsidy"
    documentation = "Reduction in energy bills due to offsetting the price cap and compensating energy firms."
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        total_subsidy = 0
        for month in range(1, MONTHS_IN_YEAR + 1):
            total_subsidy = total_subsidy + household(
                "monthly_epg_subsidy", f"{period.this_year}-{month:02d}"
            )
        return total_subsidy



from policyengine_uk.model_api import *

"""
This file contains variables that are commonly used in benefit eligibility calculations.
"""


class family_benefits(Variable):
    value_type = float
    entity = Person
    label = "Total simulated family benefits for this person"
    definition_period = YEAR
    unit = GBP
    defined_for = "is_benunit_head"

    adds = [
        "child_benefit",
        "ESA_income",
        "housing_benefit",
        "income_support",
        "JSA_income",
        "pension_credit",
        "universal_credit",
        "council_tax_benefit",
        "child_tax_credit",
        "working_tax_credit",
    ]


class family_benefits_reported(Variable):
    value_type = float
    entity = Person
    label = "Total reported family benefits for this person"
    definition_period = YEAR
    unit = GBP

    # Currently no family benefits reported.


class benefits(Variable):
    value_type = float
    entity = Person
    label = "Total benefits"
    documentation = "Total state benefits"
    unit = GBP
    definition_period = YEAR

    def formula(person, period, parameters):
        return add(person, period, ["personal_benefits", "family_benefits"])


class household_benefits(Variable):
    value_type = float
    entity = Household
    label = "benefits"
    documentation = "Total value of benefits received by household"
    definition_period = YEAR
    unit = GBP
    adds = [
        "child_benefit",
        "ESA_income",
        "housing_benefit",
        "income_support",
        "JSA_income",
        "pension_credit",
        "universal_credit",
        "working_tax_credit",
        "child_tax_credit",
        "attendance_allowance",
        "AFCS",
        "BSP",
        "carers_allowance",
        "dla",
        "ESA_contrib",
        "IIDB",
        "incapacity_benefit",
        "JSA_contrib",
        "pip",
        "sda",
        "state_pension",
        "maternity_allowance",
        "SSP",
        "SMP",
        "ssmg",
        "basic_income",
        "epg_subsidy",
    ]

    def formula(household, period, parameters):
        contrib = parameters(period).gov.contrib
        uprating = contrib.benefit_uprating
        benefits = household_benefits.adds
        if contrib.abolish_council_tax:
            benefits = [
                benefit
                for benefit in benefits
                if benefit != "council_tax_benefit"
            ]
        general_benefits = add(
            household,
            period,
            [
                benefit
                for benefit in benefits
                if benefit not in ["basic_income"]
            ],
        )
        non_sp_benefits = add(
            household,
            period,
            [
                benefit
                for benefit in benefits
                if benefit not in ["state_pension", "basic_income"]
            ],
        )
        return (
            add(household, period, benefits)
            + general_benefits * uprating.all
            + non_sp_benefits * uprating.non_sp
        )


class other_benefits(Variable):
    value_type = float
    entity = Person
    label = "Income from benefits not modelled or detailed in the model"
    definition_period = YEAR
    unit = GBP

    adds = ["benefits_reported"]
    subtracts = ["personal_benefits", "family_benefits"]


class benefits_reported(Variable):
    value_type = float
    entity = Person
    label = "Total simulated"
    definition_period = YEAR
    unit = GBP

    adds = ["personal_benefits_reported", "family_benefits_reported"]


class benefits_modelling(Variable):
    value_type = float
    entity = Person
    label = (
        "Difference between reported and simulated benefits for this person"
    )
    definition_period = YEAR
    unit = GBP

    adds = ["benefits"]
    subtracts = ["benefits_reported"]


class is_QYP(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is a qualifying young person for benefits purposes"
    definition_period = YEAR

    def formula(person, period, parameters):
        education = person("current_education", period)
        under_20 = person("age", period) < 20
        in_education = ~(
            education == education.possible_values.NOT_IN_EDUCATION
        )
        return under_20 & in_education


class is_child_or_QYP(Variable):
    value_type = bool
    entity = Person
    label = "Whether this person is a child or qualifying young person for most benefits"
    definition_period = YEAR

    def formula(person, period, parameters):
        return (person("age", period) < 16) | person("is_QYP", period)


class benefits_premiums(Variable):
    value_type = float
    entity = BenUnit
    label = "Value of premiums for disability and carer status"
    definition_period = YEAR
    unit = GBP

    adds = [
        "disability_premium",
        "enhanced_disability_premium",
        "severe_disability_premium",
        "carer_premium",
    ]


class benunit_weekly_hours(Variable):
    value_type = float
    entity = BenUnit
    label = "Average weekly hours worked by adults in the benefit unit"
    definition_period = YEAR
    unit = "hour"

    adds = ["weekly_hours"]


class is_single(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether this benefit unit contains a single claimant for benefits"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        relation_type = benunit("relation_type", period)
        relations = relation_type.possible_values
        return relation_type == relations.SINGLE


class is_couple(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether this benefit unit contains a joint couple claimant for benefits"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        relation_type = benunit("relation_type", period)
        relations = relation_type.possible_values
        return relation_type == relations.COUPLE


class is_lone_parent(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether the family is a lone parent family"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        return family_type == families.LONE_PARENT


class is_single_person(Variable):
    value_type = bool
    entity = BenUnit
    label = "Whether the family is a single person"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        family_type = benunit("family_type", period)
        families = family_type.possible_values
        return family_type == families.SINGLE


class personal_benefits(Variable):
    value_type = float
    entity = Person
    label = "Value of personal, non-means-tested benefits"
    definition_period = YEAR
    unit = GBP

    adds = [
        "attendance_allowance",
        "AFCS",
        "BSP",
        "carers_allowance",
        "dla",
        "ESA_contrib",
        "IIDB",
        "incapacity_benefit",
        "JSA_contrib",
        "pip",
        "sda",
        "state_pension",
        "student_payments",
        "student_loans",
        "maternity_allowance",
        "SSP",
        "SMP",
        "ssmg",
        "basic_income",
    ]


class personal_benefits_reported(Variable):
    value_type = float
    entity = Person
    label = "Value of personal, non-means-tested benefits"
    definition_period = YEAR
    unit = GBP

    adds = [
        i + "_reported"
        for i in [
            "AA",
            "AFCS",
            "BSP",
            "carers_allowance",
            "DLA_M",
            "DLA_SC",
            "ESA_contrib",
            "IIDB",
            "incapacity_benefit",
            "JSA_contrib",
            "PIP_M",
            "PIP_DL",
            "SDA",
            "state_pension",
        ]
    ]


class claims_all_entitled_benefits(Variable):
    value_type = bool
    entity = BenUnit
    label = "Claims all eligible benefits"
    definition_period = YEAR
    documentation = (
        "Whether this family would claim any benefit they are entitled to"
    )

    def formula(benunit, period, parameters):
        # Return false we have any reported values in the simulation for benefits.
        return (
            add(
                benunit,
                period,
                [
                    "child_tax_credit_reported",
                    "working_tax_credit_reported",
                    "universal_credit_reported",
                    "housing_benefit_reported",
                    "JSA_income_reported",
                    "income_support_reported",
                    "ESA_income_reported",
                ],
            ).sum()
            < 1
        )


class claims_legacy_benefits(Variable):
    value_type = bool
    entity = BenUnit
    label = "Claims legacy benefits"
    documentation = "Whether this family is currently receiving legacy benefits (overrides UC claimant status)"
    definition_period = YEAR

    def formula(benunit, period, parameters):
        BENEFITS = [
            "child_tax_credit",
            "working_tax_credit",
            "housing_benefit_pre_benefit_cap",
            "ESA_income",
            "income_support",
            "JSA_income",
        ]

        return add(benunit, period, BENEFITS) > 0



from policyengine_uk.model_api import *


class pension_credit(Variable):
    label = "Pension Credit"
    entity = BenUnit
    definition_period = YEAR
    value_type = float
    unit = GBP
    reference = "https://www.legislation.gov.uk/ukpga/2002/16/contents"
    defined_for = "would_claim_pc"
    adds = ["pension_credit_entitlement"]



from policyengine_uk.model_api import *


class full_rate_vat_consumption(Variable):
    label = "consumption of VAT full-rated goods and services"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        # If unknown, assume half of consumption is VAT full-rated.
        return household("consumption", period) * 0.5


class reduced_rate_vat_consumption(Variable):
    label = "consumption of VAT reduced-rated goods and services"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = "currency-GBP"

    def formula(household, period, parameters):
        vat = parameters(period).gov.hmrc.vat
        consumption = household("consumption", period)
        return consumption * vat.reduced_rate_share



from policyengine_uk.model_api import *


class shareholding(Variable):
    label = "Share in the corporate sector"
    documentation = "Exposure to taxes on corporations"
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    def formula(household, period, parameters):
        wealth = parameters(period).household.wealth
        nbs = wealth.national_balance_sheet
        wealth = household("corporate_wealth", period)
        total_wealth = nbs.household.financial_net_worth
        return wealth / total_wealth


class corporate_tax_incidence(Variable):
    label = "Corporate tax incidence"
    documentation = (
        "Reduction in value of corporate wealth due to taxes on corporations"
    )
    entity = Household
    definition_period = YEAR
    value_type = float
    unit = GBP

    adds = [
        "corporate_sdlt_change_incidence",
        "business_rates_change_incidence",
    ]



from policyengine_uk.model_api import *


class basic_income(Variable):
    label = "Basic income"
    entity = Person
    definition_period = YEAR
    value_type = float
    unit = GBP

    adds = ["bi_maximum"]
    subtracts = ["bi_phaseout"]



from policyengine_uk.model_api import *


class AFCS(Variable):
    value_type = float
    entity = Person
    label = "Armed Forces Compensation Scheme"
    definition_period = YEAR
    unit = GBP
    uprating = "calibration.uprating.september_cpi"

    adds = ["AFCS_reported"]


class AFCS_reported(Variable):
    value_type = float
    entity = Person
    label = "Armed Forces Compensation Scheme (reported)"
    definition_period = YEAR
    unit = GBP
